@@ -12,6 +12,7 @@ Spyder kernel for Jupyter.
 
 # Standard library imports
 from distutils.version import LooseVersion
+import logging
 import os
 import sys
 import threading
@@ -19,17 +20,23 @@ import threading
 # Third-party imports
 import ipykernel
 from ipykernel.ipkernel import IPythonKernel
-from ipykernel.zmqshell import ZMQInteractiveShell
 from traitlets.config.loader import LazyConfigValue
 
 # Local imports
-from spyder_kernels.py3compat import TEXT_TYPES, to_text_string
-from spyder_kernels.comms.frontendcomm import FrontendComm
-from spyder_kernels.py3compat import PY3, input
+from spyder_kernels.py3compat import (
+    TEXT_TYPES, to_text_string, PY3, input, TimeoutError)
+from spyder_kernels.comms.frontendcomm import FrontendComm, CommError
 from spyder_kernels.utils.iofuncs import iofunctions
 from spyder_kernels.utils.mpl import (
     MPL_BACKENDS_FROM_SPYDER, MPL_BACKENDS_TO_SPYDER, INLINE_FIGURE_FORMATS)
 from spyder_kernels.utils.nsview import get_remote_data, make_remote_view
+from spyder_kernels.console.shell import SpyderShell
+
+if PY3:
+    import faulthandler
+
+
+logger = logging.getLogger(__name__)
 
 
 # Excluded variables from the Variable Explorer (i.e. they are not
@@ -37,27 +44,6 @@ from spyder_kernels.utils.nsview import get_remote_data, make_remote_view
 EXCLUDED_NAMES = ['In', 'Out', 'exit', 'get_ipython', 'quit']
 
 
-class SpyderShell(ZMQInteractiveShell):
-    """Spyder shell."""
-
-    def ask_exit(self):
-        """Engage the exit actions."""
-        self.kernel.frontend_comm.close_thread()
-        return super(SpyderShell, self).ask_exit()
-
-    def get_local_scope(self, stack_depth):
-        """Get local scope at given frame depth."""
-        frame = sys._getframe(stack_depth + 1)
-        if self.kernel._pdb_frame is frame:
-            # we also give the globals because they might not be in
-            # self.user_ns
-            namespace = frame.f_globals.copy()
-            namespace.update(self.kernel._pdb_locals)
-            return namespace
-        else:
-            return frame.f_locals
-
-
 class SpyderKernel(IPythonKernel):
     """Spyder kernel for Jupyter."""
 
@@ -97,21 +83,28 @@ class SpyderKernel(IPythonKernel):
             'update_syspath': self.update_syspath,
             'is_special_kernel_valid': self.is_special_kernel_valid,
             'get_matplotlib_backend': self.get_matplotlib_backend,
+            'get_mpl_interactive_backend': self.get_mpl_interactive_backend,
             'pdb_input_reply': self.pdb_input_reply,
             '_interrupt_eventloop': self._interrupt_eventloop,
+            'enable_faulthandler': self.enable_faulthandler,
+            "flush_std": self.flush_std,
             }
         for call_id in handlers:
             self.frontend_comm.register_call_handler(
                 call_id, handlers[call_id])
 
         self.namespace_view_settings = {}
-        self._pdb_obj = None
-        self._pdb_step = None
         self._mpl_backend_error = None
         self._running_namespace = None
         self._pdb_input_line = None
+        self.faulthandler_handle = None
 
     # -- Public API -----------------------------------------------------------
+    def do_shutdown(self, restart):
+        """Disable faulthandler if enabled before proceeding."""
+        self.disable_faulthandler()
+        super(SpyderKernel, self).do_shutdown(restart)
+
     def frontend_call(self, blocking=False, broadcast=True,
                       timeout=None, callback=None):
         """Call the frontend."""
@@ -127,6 +120,42 @@ class SpyderKernel(IPythonKernel):
             callback=callback,
             timeout=timeout)
 
+    def flush_std(self):
+        """Flush C standard streams."""
+        sys.__stderr__.flush()
+        sys.__stdout__.flush()
+
+    def enable_faulthandler(self, fn):
+        """
+        Open a file to save the faulthandling and identifiers for
+        internal threads.
+        """
+        if not PY3:
+            # Not implemented
+            return
+        self.disable_faulthandler()
+        f = open(fn, 'w')
+        self.faulthandler_handle = f
+        f.write("Main thread id:\n")
+        f.write(hex(threading.main_thread().ident))
+        f.write('\nSystem threads ids:\n')
+        f.write(" ".join([hex(thread.ident) for thread in threading.enumerate()
+                          if thread is not threading.main_thread()]))
+        f.write('\n')
+        faulthandler.enable(f)
+
+    def disable_faulthandler(self):
+        """
+        Cancel the faulthandling, close the file handle and remove the file.
+        """
+        if not PY3:
+            # Not implemented
+            return
+        if self.faulthandler_handle:
+            faulthandler.disable()
+            self.faulthandler_handle.close()
+            self.faulthandler_handle = None
+
     # --- For the Variable Explorer
     def set_namespace_view_settings(self, settings):
         """Set namespace_view_settings."""
@@ -233,7 +262,7 @@ class SpyderKernel(IPythonKernel):
         """
         from spyder_kernels.utils.misc import fix_reference_name
 
-        glbs = self._mglobals()
+        glbs = self.shell.user_ns
         load_func = iofunctions.load_funcs[ext]
         data, error_message = load_func(filename)
 
@@ -263,12 +292,6 @@ class SpyderKernel(IPythonKernel):
         return iofunctions.save(data, filename)
 
     # --- For Pdb
-    def is_debugging(self):
-        """
-        Check if we are currently debugging.
-        """
-        return bool(self._pdb_frame)
-
     def _do_complete(self, code, cursor_pos):
         """Call parent class do_complete"""
         return super(SpyderKernel, self).do_complete(code, cursor_pos)
@@ -279,51 +302,59 @@ class SpyderKernel(IPythonKernel):
 
         Public method of ipykernel overwritten for debugging.
         """
-        if self.is_debugging():
-            return self._pdb_obj.do_complete(code, cursor_pos)
+        if self.shell.is_debugging():
+            return self.shell.pdb_session.do_complete(code, cursor_pos)
         return self._do_complete(code, cursor_pos)
 
-    def publish_pdb_state(self):
-        """Publish Pdb state."""
-        if self._pdb_obj:
-            state = dict(namespace_view = self.get_namespace_view(),
-                         var_properties = self.get_var_properties(),
-                         step = self._pdb_step)
+    def publish_pdb_state(self, step):
+        """
+        Publish Variable Explorer state and Pdb step through
+        send_spyder_msg.
+        """
+        state = dict(
+            namespace_view=self.get_namespace_view(),
+            var_properties=self.get_var_properties(),
+            step=step
+         )
+        try:
             self.frontend_call(blocking=False).pdb_state(state)
+        except (CommError, TimeoutError):
+            logger.debug("Could not send Pdb state to the frontend.")
 
     def set_spyder_breakpoints(self, breakpoints):
         """
         Handle a message from the frontend
         """
-        if self._pdb_obj:
-            self._pdb_obj.set_spyder_breakpoints(breakpoints)
+        if self.shell.pdb_session:
+            self.shell.pdb_session.set_spyder_breakpoints(breakpoints)
 
     def set_pdb_ignore_lib(self, state):
         """
         Change the "Ignore libraries while stepping" debugger setting.
         """
-        if self._pdb_obj:
-            self._pdb_obj.pdb_ignore_lib = state
+        if self.shell.pdb_session:
+            self.shell.pdb_session.pdb_ignore_lib = state
 
     def set_pdb_execute_events(self, state):
         """
         Handle a message from the frontend
         """
-        if self._pdb_obj:
-            self._pdb_obj.pdb_execute_events = state
+        if self.shell.pdb_session:
+            self.shell.pdb_session.pdb_execute_events = state
 
     def set_pdb_use_exclamation_mark(self, state):
         """
         Set an option on the current debugging session to decide wether
         the Pdb commands needs to be prefixed by '!'
         """
-        if self._pdb_obj:
-            self._pdb_obj.pdb_use_exclamation_mark = state
+        if self.shell.pdb_session:
+            self.shell.pdb_session.pdb_use_exclamation_mark = state
 
     def pdb_input_reply(self, line, echo_stack_entry=True):
         """Get a pdb command from the frontend."""
-        if self._pdb_obj:
-            self._pdb_obj._disable_next_stack_entry = not echo_stack_entry
+        if self.shell.pdb_session:
+            self.shell.pdb_session._disable_next_stack_entry = (
+                not echo_stack_entry)
         self._pdb_input_line = line
         if self.eventloop:
             # Interrupting the eventloop is only implemented when a message is
@@ -340,7 +371,7 @@ class SpyderKernel(IPythonKernel):
         Runs the eventloop while debugging.
         """
         # Only works if the comm is open and this is a pdb prompt.
-        if not self.frontend_comm.is_open() or not self._pdb_frame:
+        if not self.frontend_comm.is_open() or not self.shell.is_debugging():
             return input(prompt)
 
         # Flush output before making the request.
@@ -418,6 +449,60 @@ class SpyderKernel(IPythonKernel):
         except Exception:
             return None
 
+    def get_mpl_interactive_backend(self):
+        """
+        Get current Matplotlib interactive backend.
+
+        This is different from the current backend because, for instance, the
+        user can set first the Qt5 backend, then the Inline one. In that case,
+        the current backend is Inline, but the current interactive one is Qt5,
+        and this backend can't be changed without a kernel restart.
+        """
+        # Mapping from frameworks to backend names.
+        mapping = {
+            'qt': 'QtAgg',  # For Matplotlib 3.5+
+            'qt5': 'Qt5Agg',
+            'tk': 'TkAgg',
+            'macosx': 'MacOSX'
+        }
+
+        try:
+            # --- Get interactive framework
+            framework = None
+
+            # This is necessary because _get_running_interactive_framework
+            # can't detect Tk in a Jupyter kernel.
+            if hasattr(self, 'app_wrapper'):
+                if hasattr(self.app_wrapper, 'app'):
+                    import tkinter
+                    if isinstance(self.app_wrapper.app, tkinter.Tk):
+                        framework = 'tk'
+
+            if framework is None:
+                try:
+                    # This is necessary for Matplotlib 3.3.0+
+                    from matplotlib import cbook
+                    framework = cbook._get_running_interactive_framework()
+                except AttributeError:
+                    # For older versions
+                    from matplotlib import backends
+                    framework = backends._get_running_interactive_framework()
+
+            # --- Return backend according to framework
+            if framework is None:
+                # Since no interactive backend has been set yet, this is
+                # equivalent to having the inline one.
+                return 0
+            elif framework in mapping:
+                return MPL_BACKENDS_TO_SPYDER[mapping[framework]]
+            else:
+                # This covers the case of other backends (e.g. Wx or Gtk)
+                # which users can set interactively with the %matplotlib
+                # magic but not through our Preferences.
+                return -1
+        except Exception:
+            return None
+
     def set_matplotlib_backend(self, backend, pylab=False):
         """Set matplotlib backend given a Spyder backend option."""
         mpl_backend = MPL_BACKENDS_FROM_SPYDER[to_text_string(backend)]
@@ -566,16 +651,16 @@ class SpyderKernel(IPythonKernel):
         """
         ns = {}
         if self._running_namespace is None:
-            ns.update(self._mglobals())
+            ns.update(self.shell.user_ns)
         else:
+            # This is true when a file is executing.
             running_globals, running_locals = self._running_namespace
             ns.update(running_globals)
             if running_locals is not None:
                 ns.update(running_locals)
 
-        if self._pdb_frame is not None:
-            ns.update(self._pdb_locals)
-
+        # Add debugging locals
+        ns.update(self.shell._pdb_locals)
         # Add magics to ns so we can show help about them on the Help
         # plugin
         if with_magics:
@@ -583,7 +668,6 @@ class SpyderKernel(IPythonKernel):
             cell_magics = self.shell.magics_manager.magics['cell']
             ns.update(line_magics)
             ns.update(cell_magics)
-
         return ns
 
     def _get_reference_namespace(self, name):
@@ -592,22 +676,10 @@ class SpyderKernel(IPythonKernel):
 
         It returns the globals() if reference has not yet been defined
         """
-        glbs = self._mglobals()
-        if self._pdb_frame is None:
-            return glbs
-        else:
-            lcls = self._pdb_locals
-            if name in lcls:
-                return lcls
-            else:
-                return glbs
-
-    def _mglobals(self):
-        """Return current globals -- handles Pdb frames"""
-        if self._pdb_frame is not None:
-            return self._pdb_frame.f_globals
-        else:
-            return self.shell.user_ns
+        lcls = self.shell._pdb_locals
+        if name in lcls:
+            return lcls
+        return self.shell.user_ns
 
     def _get_len(self, var):
         """Return sequence length"""
@@ -668,28 +740,6 @@ class SpyderKernel(IPythonKernel):
         except:
             return None
 
-    # --- For Pdb
-    def _register_pdb_session(self, pdb_obj):
-        """Register Pdb session to use it later"""
-        self._pdb_obj = pdb_obj
-
-    @property
-    def _pdb_frame(self):
-        """Return current Pdb frame if there is any"""
-        if self._pdb_obj is not None and self._pdb_obj.curframe is not None:
-            return self._pdb_obj.curframe
-
-    @property
-    def _pdb_locals(self):
-        """
-        Return current Pdb frame locals if available. Otherwise
-        return an empty dictionary
-        """
-        if self._pdb_frame:
-            return self._pdb_obj.curframe_locals
-        else:
-            return {}
-
     # --- For the Help plugin
     def _eval(self, text):
         """
@@ -718,7 +768,12 @@ class SpyderKernel(IPythonKernel):
         """
         import traceback
         from IPython.core.getipython import get_ipython
-        import matplotlib
+
+        # Don't proceed further if there's any error while importing Matplotlib
+        try:
+            import matplotlib
+        except Exception:
+            return
 
         generic_error = (
             "\n" + "="*73 + "\n"
@@ -756,6 +811,15 @@ class SpyderKernel(IPythonKernel):
             # This covers other RuntimeError's
             else:
                 error = generic_error.format(traceback.format_exc())
+        except ImportError as err:
+            additional_info = (
+                "This is most likely caused by missing packages in the Python "
+                "environment\n"
+                "or installation whose interpreter is located at:\n\n"
+                "    {0}"
+            ).format(sys.executable)
+
+            error = generic_error.format(err) + '\n\n' + additional_info
         except Exception:
             error = generic_error.format(traceback.format_exc())
 
